/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
var __classPrivateFieldGet = (this && this.__classPrivateFieldGet) || function (receiver, state, kind, f) {
    if (kind === "a" && !f) throw new TypeError("Private accessor was defined without a getter");
    if (typeof state === "function" ? receiver !== state || !f : !state.has(receiver)) throw new TypeError("Cannot read private member from an object whose class did not declare it");
    return kind === "m" ? f : kind === "a" ? f.call(receiver) : f ? f.value : state.get(receiver);
};
var _SharedTreeBranch_events;
import { assert, oob } from "@fluidframework/core-utils/internal";
import { measure } from "@fluidframework/telemetry-utils/internal";
import { CommitKind, findAncestor, makeAnonChange, mintCommit, rebaseBranch, tagRollbackInverse, } from "../core/index.js";
import { createEmitter } from "../events/index.js";
import { TransactionStack } from "./transactionStack.js";
import { fail } from "../util/index.js";
/**
 * Returns the operation that caused the given {@link SharedTreeBranchChange}.
 */
export function getChangeReplaceType(change) {
    // The "replace" variant of the change event is emitted by two operations: committing a transaction and doing a rebase.
    // Committing a transaction will always remove one or more commits (the commits that were squashed),
    // and will add exactly one new commit (the squash commit).
    if (change.removedCommits.length === 0 || change.newCommits.length !== 1) {
        return "rebase";
    }
    // There is only one case in which a rebase both removes commits and adds exactly one new commit.
    // This occurs when there is exactly one divergent, but equivalent, commit on each branch:
    //
    // A ─ B (branch X)	  -- rebase Y onto X -->   A ─ B (branch X)
    // └─ B' (branch Y)                                └─ (branch Y)
    //
    // B' is removed and replaced by B because both have the same revision.
    assert(change.removedCommits[0] !== undefined, 0x9e4 /* This wont run due to the length check above */);
    assert(change.newCommits[0] !== undefined, 0x9e5 /* This wont run because a replace operation always has new commits */);
    if (change.removedCommits.length === 1 &&
        change.removedCommits[0].revision === change.newCommits[0].revision) {
        return "rebase";
    }
    return "transactionCommit";
}
/**
 * A branch of changes that can be applied to a SharedTree.
 */
export class SharedTreeBranch {
    /**
     * Construct a new branch.
     * @param head - the head of the branch
     * @param changeFamily - determines the set of changes that this branch can commit
     * @param branchTrimmer - an optional event emitter that informs the branch it has been trimmed. If this is not supplied, then the branch must
     * never be trimmed. See {@link BranchTrimmingEvents} for details on trimming.
     */
    constructor(head, changeFamily, mintRevisionTag, branchTrimmer, telemetryEventBatcher) {
        this.head = head;
        this.changeFamily = changeFamily;
        this.mintRevisionTag = mintRevisionTag;
        this.branchTrimmer = branchTrimmer;
        this.telemetryEventBatcher = telemetryEventBatcher;
        _SharedTreeBranch_events.set(this, createEmitter());
        this.events = __classPrivateFieldGet(this, _SharedTreeBranch_events, "f");
        this.transactions = new TransactionStack();
        /**
         * After pushing a starting revision to the transaction stack, this branch might be rebased
         * over commits which are children of that starting revision. When the transaction is committed,
         * those rebased-over commits should not be included in the transaction's squash commit, even though
         * they exist between the starting revision and the final commit within the transaction.
         *
         * Whenever `rebaseOnto` is called during a transaction, this map is augmented with an entry from the
         * original merge-base to the new merge-base.
         *
         * This state need only be retained for the lifetime of the transaction.
         *
         * TODO: This strategy might need to be revisited when adding better support for async transactions.
         * Since:
         *
         * 1. Transactionality is guaranteed primarily by squashing at commit time
         * 2. Branches may be rebased with an ongoing transaction
         *
         * a rebase operation might invalidate only a portion of a transaction's commits, thus defeating the
         * purpose of transactionality.
         *
         * AB#6483 and children items track this work.
         */
        this.initialTransactionRevToRebasedRev = new Map();
        this.disposed = false;
        this.editor = this.changeFamily.buildEditor(mintRevisionTag, (change) => this.apply(change));
        this.unsubscribeBranchTrimmer = branchTrimmer?.on("ancestryTrimmed", (commit) => {
            __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("ancestryTrimmed", commit);
        });
    }
    /**
     * Sets the head of this branch. Emits no change events.
     */
    setHead(head) {
        this.assertNotDisposed();
        assert(!this.isTransacting(), 0x685 /* Cannot set head during a transaction */);
        this.head = head;
    }
    /**
     * Apply a change to this branch.
     * @param taggedChange - the change to apply
     * @param kind - the kind of change to apply
     * @returns the change that was applied and the new head commit of the branch
     */
    apply(taggedChange, kind = CommitKind.Default) {
        this.assertNotDisposed();
        const revisionTag = taggedChange.revision;
        assert(revisionTag !== undefined, 0xa49 /* Revision tag must be provided */);
        const newHead = mintCommit(this.head, {
            revision: revisionTag,
            change: taggedChange.change,
        });
        const changeEvent = {
            type: "append",
            kind,
            change: taggedChange,
            newCommits: [newHead],
        };
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("beforeChange", changeEvent);
        this.head = newHead;
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("afterChange", changeEvent);
        return [taggedChange.change, newHead];
    }
    /**
     * @returns the commit at the head of this branch.
     */
    getHead() {
        return this.head;
    }
    /**
     * Begin a transaction on this branch. If the transaction is committed via {@link commitTransaction},
     * all commits made since this call will be squashed into a single head commit.
     */
    startTransaction() {
        this.assertNotDisposed();
        const forks = new Set();
        const onDisposeUnSubscribes = [];
        const onForkUnSubscribe = onForkTransitive(this, (fork) => {
            forks.add(fork);
            onDisposeUnSubscribes.push(fork.events.on("dispose", () => forks.delete(fork)));
        });
        this.transactions.push(this.head.revision, () => {
            forks.forEach((fork) => fork.dispose());
            onDisposeUnSubscribes.forEach((unsubscribe) => unsubscribe());
            onForkUnSubscribe();
        });
        this.editor.enterTransaction();
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("transactionStarted", this.transactions.size === 1);
    }
    /**
     * Commit the current transaction. There must be a transaction in progress that was begun via {@link startTransaction}.
     * If there are commits in the current transaction, they will be squashed into a new single head commit.
     * @returns the commits that were squashed and the new squash commit if a squash occurred, otherwise `undefined`.
     * @remarks If the transaction had no changes applied during its lifetime, then no squash occurs (i.e. this method is a no-op).
     * Even if the transaction contained only one change, it will still be replaced with an (equivalent) squash change.
     */
    commitTransaction() {
        this.assertNotDisposed();
        const [startCommit, commits] = this.popTransaction();
        this.editor.exitTransaction();
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("transactionCommitted", this.transactions.size === 0);
        if (commits.length === 0) {
            return undefined;
        }
        // Squash the changes and make the squash commit the new head of this branch
        const squashedChange = this.changeFamily.rebaser.compose(commits);
        const revision = this.mintRevisionTag();
        const newHead = mintCommit(startCommit, {
            revision,
            change: this.changeFamily.rebaser.changeRevision(squashedChange, revision),
        });
        const changeEvent = {
            type: "replace",
            change: undefined,
            removedCommits: commits,
            newCommits: [newHead],
        };
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("beforeChange", changeEvent);
        this.head = newHead;
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("afterChange", changeEvent);
        return [commits, newHead];
    }
    /**
     * Cancel the current transaction. There must be a transaction in progress that was begun via
     * {@link startTransaction}. All commits made during the transaction will be removed.
     * @returns the change to this branch resulting in the removal of the commits, and a list of the
     * commits that were removed.
     */
    abortTransaction() {
        this.assertNotDisposed();
        const [startCommit, commits] = this.popTransaction();
        this.editor.exitTransaction();
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("transactionAborted", this.transactions.size === 0);
        if (commits.length === 0) {
            __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("transactionRolledBack", this.transactions.size === 0);
            return [undefined, []];
        }
        const inverses = [];
        for (let i = commits.length - 1; i >= 0; i--) {
            const revision = this.mintRevisionTag();
            const commit = commits[i] ?? fail("This wont run because we are iterating through commits");
            const inverse = this.changeFamily.rebaser.changeRevision(this.changeFamily.rebaser.invert(commit, true, revision), revision, commit.revision);
            inverses.push(tagRollbackInverse(inverse, revision, commit.revision));
        }
        const change = inverses.length > 0 ? this.changeFamily.rebaser.compose(inverses) : undefined;
        const changeEvent = {
            type: "remove",
            change: change === undefined ? undefined : makeAnonChange(change),
            removedCommits: commits,
        };
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("beforeChange", changeEvent);
        this.head = startCommit;
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("afterChange", changeEvent);
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("transactionRolledBack", this.transactions.size === 0);
        return [change, commits];
    }
    /**
     * True iff this branch is in the middle of a transaction that was begin via {@link startTransaction}
     */
    isTransacting() {
        return this.transactions.size !== 0;
    }
    popTransaction() {
        const { startRevision: startRevisionOriginal } = this.transactions.pop();
        let startRevision = startRevisionOriginal;
        while (this.initialTransactionRevToRebasedRev.has(startRevision)) {
            startRevision = this.initialTransactionRevToRebasedRev.get(startRevision) ?? oob();
        }
        if (!this.isTransacting()) {
            this.initialTransactionRevToRebasedRev.clear();
        }
        const commits = [];
        const startCommit = findAncestor([this.head, commits], (c) => c.revision === startRevision);
        assert(startCommit !== undefined, 0x593 /* Expected branch to be ahead of transaction start revision */);
        return [startCommit, commits];
    }
    /**
     * Spawn a new branch that is based off of the current state of this branch.
     * @param commit - The commit to base the new branch off of. Defaults to the head of this branch.
     * @remarks Changes made to the new branch will not be applied to this branch until the new branch is {@link SharedTreeBranch.merge | merged} back in.
     * Forks created during a transaction will be disposed when the transaction ends.
     */
    fork(commit = this.head) {
        this.assertNotDisposed();
        const fork = new SharedTreeBranch(commit, this.changeFamily, this.mintRevisionTag, this.branchTrimmer);
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("fork", fork);
        return fork;
    }
    /**
     * Rebase the changes that have been applied to this branch over divergent changes in the given branch.
     * After this operation completes, this branch will be based off of `branch`.
     *
     * @param branch - the branch to rebase onto
     * @param upTo - the furthest commit on `branch` over which to rebase (inclusive). Defaults to the head commit of `branch`.
     * @returns the result of the rebase or undefined if nothing changed
     */
    rebaseOnto(branch, upTo = branch.getHead()) {
        this.assertNotDisposed();
        // Rebase this branch onto the given branch
        const rebaseResult = this.rebaseBranch(this, branch, upTo);
        if (rebaseResult === undefined) {
            return undefined;
        }
        // The net change to this branch is provided by the `rebaseBranch` API
        const { newSourceHead, commits } = rebaseResult;
        const { deletedSourceCommits, targetCommits, sourceCommits } = commits;
        const newCommits = targetCommits.concat(sourceCommits);
        if (this.isTransacting()) {
            const firstCommit = targetCommits[0] ?? oob();
            const lastCommit = targetCommits[targetCommits.length - 1] ?? oob();
            const src = firstCommit.parent?.revision;
            const dst = lastCommit.revision;
            if (src !== undefined && dst !== undefined) {
                this.initialTransactionRevToRebasedRev.set(src, dst);
            }
        }
        const changeEvent = {
            type: "replace",
            get change() {
                const change = rebaseResult.sourceChange;
                return change === undefined ? undefined : makeAnonChange(change);
            },
            removedCommits: deletedSourceCommits,
            newCommits,
        };
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("beforeChange", changeEvent);
        this.head = newSourceHead;
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("afterChange", changeEvent);
        return rebaseResult;
    }
    /**
     * Apply all the divergent changes on the given branch to this branch.
     *
     * @param branch - the branch to merge into this branch
     * @returns the net change to this branch and the commits that were added to this branch by the merge,
     * or undefined if nothing changed
     */
    merge(branch) {
        this.assertNotDisposed();
        branch.assertNotDisposed();
        assert(!branch.isTransacting(), 0x597 /* Branch may not be merged while transaction is in progress */);
        if (branch === this) {
            return undefined;
        }
        // Rebase the given branch onto this branch
        const rebaseResult = this.rebaseBranch(branch, this);
        if (rebaseResult === undefined) {
            return undefined;
        }
        // Compute the net change to this branch
        const sourceCommits = rebaseResult.commits.sourceCommits;
        const change = this.changeFamily.rebaser.compose(sourceCommits);
        const taggedChange = makeAnonChange(change);
        const changeEvent = {
            type: "append",
            kind: CommitKind.Default,
            get change() {
                return taggedChange;
            },
            newCommits: sourceCommits,
        };
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("beforeChange", changeEvent);
        this.head = rebaseResult.newSourceHead;
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("afterChange", changeEvent);
        return [change, sourceCommits];
    }
    /** Rebase `branchHead` onto `onto`, but return undefined if nothing changed */
    rebaseBranch(branch, onto, upTo = onto.getHead()) {
        const { head } = branch;
        if (head === upTo) {
            return undefined;
        }
        const { duration, output } = measure(() => rebaseBranch(this.mintRevisionTag, this.changeFamily.rebaser, head, upTo, onto.getHead()));
        this.telemetryEventBatcher?.accumulateAndLog({ duration, ...output.telemetryProperties });
        if (this.head === output.newSourceHead) {
            return undefined;
        }
        return output;
    }
    /**
     * Dispose this branch, freezing its state.
     *
     * @remarks
     * Attempts to further mutate the branch will error.
     * Any transactions in progress will be aborted.
     * Calling dispose more than once has no effect.
     */
    dispose() {
        if (this.disposed) {
            return;
        }
        while (this.isTransacting()) {
            this.abortTransaction();
        }
        this.unsubscribeBranchTrimmer?.();
        this.disposed = true;
        __classPrivateFieldGet(this, _SharedTreeBranch_events, "f").emit("dispose");
    }
    assertNotDisposed() {
        assert(!this.disposed, 0x66e /* Branch is disposed */);
    }
}
_SharedTreeBranch_events = new WeakMap();
/**
 * Registers an event listener that fires when the given branch forks.
 * The listener will also fire when any of those forks fork, and when those forks of forks fork, and so on.
 * @param branch - the branch that will be listened to for forks
 * @param onFork - the fork event listener
 * @returns a function which when called will deregister all registrations (including transitive) created by this function.
 * The deregister function has undefined behavior if called more than once.
 */
// Branches are invariant over TChange
// eslint-disable-next-line @typescript-eslint/no-explicit-any
export function onForkTransitive(branch, onFork) {
    const offs = [];
    offs.push(branch.events.on("fork", (fork) => {
        offs.push(onForkTransitive(fork, onFork));
        onFork(fork);
    }));
    return () => offs.forEach((off) => off());
}
//# sourceMappingURL=branch.js.map