/*!
 * Copyright (c) Microsoft Corporation and contributors. All rights reserved.
 * Licensed under the MIT License.
 */
import { assert, oob } from "@fluidframework/core-utils/internal";
import { UsageError, } from "@fluidframework/telemetry-utils/internal";
import { noopValidator } from "../codec/index.js";
import { AnchorSet, CommitKind, RevertibleStatus, TreeStoredSchemaRepository, combineVisitors, makeDetachedFieldIndex, rebaseChange, rootFieldKey, tagChange, visitDelta, } from "../core/index.js";
import { createEmitter, } from "../events/index.js";
import { buildForest, createNodeKeyManager, intoDelta, jsonableTreeFromCursor, makeFieldBatchCodec, } from "../feature-libraries/index.js";
import { SharedTreeBranch, getChangeReplaceType, } from "../shared-tree-core/index.js";
import { Breakable, TransactionResult, disposeSymbol, fail } from "../util/index.js";
import { SharedTreeChangeFamily, hasSchemaChange } from "./sharedTreeChangeFamily.js";
import { SchematizingSimpleTreeView } from "./schematizingTreeView.js";
/**
 * Creates a {@link TreeCheckout}.
 * @param args - an object containing optional components that will be used to build the view.
 * Any components not provided will be created by default.
 * @remarks This does not create a {@link SharedTree}, but rather a view with the minimal state
 * and functionality required to implement {@link ITreeCheckout}.
 */
export function createTreeCheckout(idCompressor, mintRevisionTag, revisionTagCodec, args) {
    const forest = args?.forest ?? buildForest();
    const schema = args?.schema ?? new TreeStoredSchemaRepository();
    const defaultCodecOptions = { jsonValidator: noopValidator };
    const defaultFieldBatchVersion = 1;
    const changeFamily = args?.changeFamily ??
        new SharedTreeChangeFamily(revisionTagCodec, args?.fieldBatchCodec ??
            makeFieldBatchCodec(defaultCodecOptions, defaultFieldBatchVersion), { jsonValidator: noopValidator }, args?.chunkCompressionStrategy, idCompressor);
    const branch = args?.branch ??
        new SharedTreeBranch({
            change: changeFamily.rebaser.compose([]),
            revision: "root",
        }, changeFamily, () => idCompressor.generateCompressedId());
    const events = args?.events ?? createEmitter();
    const transaction = new Transaction(branch);
    return new TreeCheckout(transaction, branch, false, changeFamily, schema, forest, events, mintRevisionTag, revisionTagCodec, idCompressor, args?.removedRoots, args?.logger, args?.breaker);
}
class Transaction {
    constructor(branch) {
        this.branch = branch;
    }
    start() {
        this.branch.startTransaction();
        this.branch.editor.enterTransaction();
    }
    commit() {
        this.branch.commitTransaction();
        this.branch.editor.exitTransaction();
        return TransactionResult.Commit;
    }
    abort() {
        this.branch.abortTransaction();
        this.branch.editor.exitTransaction();
        return TransactionResult.Abort;
    }
    inProgress() {
        return this.branch.isTransacting();
    }
}
/**
 * An implementation of {@link ITreeCheckoutFork}.
 */
export class TreeCheckout {
    constructor(transaction, _branch, 
    /** True if and only if this checkout is for a forked branch and not the "main branch" of the tree. */
    isBranch, changeFamily, storedSchema, forest, events, mintRevisionTag, revisionTagCodec, idCompressor, removedRoots = makeDetachedFieldIndex("repair", revisionTagCodec, idCompressor), 
    /** Optional logger for telemetry. */
    logger, breaker = new Breakable("TreeCheckout")) {
        this.transaction = transaction;
        this._branch = _branch;
        this.isBranch = isBranch;
        this.changeFamily = changeFamily;
        this.storedSchema = storedSchema;
        this.forest = forest;
        this.events = events;
        this.mintRevisionTag = mintRevisionTag;
        this.revisionTagCodec = revisionTagCodec;
        this.idCompressor = idCompressor;
        this.removedRoots = removedRoots;
        this.logger = logger;
        this.breaker = breaker;
        this.disposed = false;
        this.views = new Set();
        /**
         * Set of revertibles maintained for automatic disposal
         */
        this.revertibles = new Set();
        /**
         * Each branch's head commit corresponds to a revertible commit.
         * Maintaining a whole branch ensures the commit graph is not pruned in a way that would prevent the commit from
         * being reverted.
         */
        this.revertibleCommitBranches = new Map();
        /**
         * copies of the removed roots used as snapshots for reverting to previous state when transactions are aborted
         */
        this.removedRootsSnapshots = [];
        // when a transaction is started, take a snapshot of the current state of removed roots
        _branch.events.on("transactionStarted", () => {
            this.removedRootsSnapshots.push(this.removedRoots.clone());
        });
        // when a transaction is committed, the latest snapshot of removed roots can be discarded
        _branch.events.on("transactionCommitted", () => {
            this.removedRootsSnapshots.pop();
        });
        // after a transaction is rolled back, revert removed roots back to the latest snapshot
        _branch.events.on("transactionRolledBack", () => {
            const snapshot = this.removedRootsSnapshots.pop();
            assert(snapshot !== undefined, 0x9ae /* a snapshot for removed roots does not exist */);
            this.removedRoots = snapshot;
        });
        // We subscribe to `beforeChange` rather than `afterChange` here because it's possible that the change is invalid WRT our forest.
        // For example, a bug in the editor might produce a malformed change object and thus applying the change to the forest will throw an error.
        // In such a case we will crash here, preventing the change from being added to the commit graph, and preventing `afterChange` from firing.
        // One important consequence of this is that we will not submit the op containing the invalid change, since op submissions happens in response to `afterChange`.
        _branch.events.on("beforeChange", (event) => {
            if (event.change !== undefined) {
                const revision = event.type === "replace"
                    ? // Change events will always contain new commits
                        // eslint-disable-next-line @typescript-eslint/no-non-null-assertion
                        event.newCommits[event.newCommits.length - 1].revision
                    : event.change.revision;
                // Conflicts due to schema will be empty and thus are not applied.
                for (const change of event.change.change.changes) {
                    if (change.type === "data") {
                        const delta = intoDelta(tagChange(change.innerChange, revision));
                        this.withCombinedVisitor((visitor) => {
                            visitDelta(delta, visitor, this.removedRoots, revision);
                        });
                    }
                    else if (change.type === "schema") {
                        // Schema changes from a current to a new schema are expected to be backwards compatible.
                        // This guarantees that all data in the forest (which is valid before the schema change)
                        // is also valid under the new schema.
                        // Note however, that such schema changes may in some cases be rolled back:
                        // Case 1: A transaction with a schema change may be aborted.
                        // The transaction may have made some data changes that would render some trees invalid
                        // under the old schema, but these changes will also be rolled back, thereby putting the forest
                        // back in the state before the transaction, which is valid under the original (reinstated) schema.
                        // Case 2: A branch with a schema change may be rebased such that the schema change (because
                        // of a constraint) is no longer applied.
                        // Such a branch may contain data changes that would render some trees invalid under the
                        // original schema. These data changes may not necessarily be rolled back.
                        // They will however be rebased over the rollback of the schema change. This rebasing will
                        // ensure that these data changes are muted if they would render some trees invalid under the
                        // original (reinstated) schema.
                        storedSchema.apply(change.innerChange.schema.new);
                    }
                    else {
                        fail("Unknown Shared Tree change type.");
                    }
                }
                this.events.emit("afterBatch");
            }
            if (event.type === "replace" && getChangeReplaceType(event) === "transactionCommit") {
                const firstCommit = event.newCommits[0] ?? oob();
                const transactionRevision = firstCommit.revision;
                for (const transactionStep of event.removedCommits) {
                    this.removedRoots.updateMajor(transactionStep.revision, transactionRevision);
                }
            }
        });
        _branch.events.on("afterChange", (event) => {
            // The following logic allows revertibles to be generated for the change.
            // Currently only appends (including merges) and transaction commits are supported.
            if (!_branch.isTransacting()) {
                if (event.type === "append" ||
                    (event.type === "replace" && getChangeReplaceType(event) === "transactionCommit")) {
                    // TODO:#20949: When the SharedTree is detached, these commits will already have been garbage collected.
                    //       Figure out a way to generate revertibles before the commits are garbage collected.
                    for (const commit of event.newCommits) {
                        const kind = event.type === "append" ? event.kind : CommitKind.Default;
                        const { change, revision } = commit;
                        const getRevertible = hasSchemaChange(change)
                            ? undefined
                            : (onRevertibleDisposed) => {
                                if (!withinEventContext) {
                                    throw new UsageError("Cannot get a revertible outside of the context of a changed event.");
                                }
                                if (this.revertibleCommitBranches.get(revision) !== undefined) {
                                    throw new UsageError("Cannot generate the same revertible more than once. Note that this can happen when multiple changed event listeners are registered.");
                                }
                                const revertibleCommits = this.revertibleCommitBranches;
                                const revertible = {
                                    get status() {
                                        const revertibleCommit = revertibleCommits.get(revision);
                                        return revertibleCommit === undefined
                                            ? RevertibleStatus.Disposed
                                            : RevertibleStatus.Valid;
                                    },
                                    revert: (release = true) => {
                                        if (revertible.status === RevertibleStatus.Disposed) {
                                            throw new UsageError("Unable to revert a revertible that has been disposed.");
                                        }
                                        const revertMetrics = this.revertRevertible(revision, kind);
                                        this.logger?.sendTelemetryEvent({
                                            eventName: TreeCheckout.revertTelemetryEventName,
                                            ...revertMetrics,
                                        });
                                        if (release) {
                                            revertible.dispose();
                                        }
                                    },
                                    dispose: () => {
                                        if (revertible.status === RevertibleStatus.Disposed) {
                                            throw new UsageError("Unable to dispose a revertible that has already been disposed.");
                                        }
                                        this.disposeRevertible(revertible, revision);
                                        onRevertibleDisposed?.(revertible);
                                    },
                                };
                                this.revertibleCommitBranches.set(revision, _branch.fork(commit));
                                this.revertibles.add(revertible);
                                return revertible;
                            };
                        let withinEventContext = true;
                        this.events.emit("changed", { isLocal: true, kind }, getRevertible);
                        withinEventContext = false;
                    }
                }
                else if (this.isRemoteChangeEvent(event)) {
                    // TODO: figure out how to plumb through commit kind info for remote changes
                    this.events.emit("changed", { isLocal: false, kind: CommitKind.Default });
                }
            }
        });
        // When the branch is trimmed, we can garbage collect any repair data whose latest relevant revision is one of the
        // trimmed revisions.
        _branch.events.on("ancestryTrimmed", (revisions) => {
            this.withCombinedVisitor((visitor) => {
                revisions.forEach((revision) => {
                    // get all the roots last created or used by the revision
                    const roots = this.removedRoots.getRootsLastTouchedByRevision(revision);
                    // get the detached field for the root and delete it from the removed roots
                    for (const root of roots) {
                        visitor.destroy(this.removedRoots.toFieldKey(root), 1);
                    }
                    this.removedRoots.deleteRootsLastTouchedByRevision(revision);
                });
            });
        });
    }
    withCombinedVisitor(fn) {
        const anchorVisitor = this.forest.anchors.acquireVisitor();
        const combinedVisitor = combineVisitors([this.forest.acquireVisitor(), anchorVisitor], [anchorVisitor]);
        fn(combinedVisitor);
        combinedVisitor.free();
    }
    checkNotDisposed(usageError) {
        if (this.disposed) {
            if (usageError !== undefined) {
                throw new UsageError(usageError);
            }
            assert(false, 0x911 /* Invalid operation on a disposed TreeCheckout */);
        }
    }
    viewWith(config) {
        const view = new SchematizingSimpleTreeView(this, config, createNodeKeyManager(this.idCompressor), this.breaker, () => {
            this.views.delete(view);
        });
        this.views.add(view);
        return view;
    }
    get rootEvents() {
        return this.forest.anchors.events;
    }
    get editor() {
        this.checkNotDisposed();
        return this._branch.editor;
    }
    locate(anchor) {
        this.checkNotDisposed();
        return this.forest.anchors.locate(anchor);
    }
    branch() {
        this.checkNotDisposed("The parent branch has already been disposed and can no longer create new branches.");
        const anchors = new AnchorSet();
        const branch = this._branch.fork();
        const storedSchema = this.storedSchema.clone();
        const forest = this.forest.clone(storedSchema, anchors);
        const transaction = new Transaction(branch);
        return new TreeCheckout(transaction, branch, true, this.changeFamily, storedSchema, forest, createEmitter(), this.mintRevisionTag, this.revisionTagCodec, this.idCompressor, this.removedRoots.clone(), this.logger, this.breaker);
    }
    rebase(checkout) {
        this.checkNotDisposed("The target of the branch rebase has been disposed and cannot be rebased.");
        checkout.checkNotDisposed("The source of the branch rebase has been disposed and cannot be rebased.");
        assert(!checkout.transaction.inProgress(), 0x9af /* A view cannot be rebased while it has a pending transaction */);
        assert(checkout.isBranch, 0xa5d /* The main branch cannot be rebased onto another branch. */);
        checkout._branch.rebaseOnto(this._branch);
    }
    rebaseOnto(checkout) {
        this.checkNotDisposed("The target of the branch rebase has been disposed and cannot be rebased.");
        checkout.rebase(this);
    }
    merge(checkout, disposeMerged = true) {
        this.checkNotDisposed("The target of the branch merge has been disposed and cannot be merged.");
        checkout.checkNotDisposed("The source of the branch merge has been disposed and cannot be merged.");
        assert(!this.transaction.inProgress(), 0x9b0 /* Views cannot be merged into a view while it has a pending transaction */);
        while (checkout.transaction.inProgress()) {
            checkout.transaction.commit();
        }
        this._branch.merge(checkout._branch);
        if (disposeMerged && checkout.isBranch) {
            // Dispose the merged checkout unless it is the main branch.
            checkout[disposeSymbol]();
        }
    }
    updateSchema(newSchema) {
        this.checkNotDisposed();
        this.editor.schema.setStoredSchema(this.storedSchema.clone(), newSchema);
    }
    dispose() {
        this[disposeSymbol]();
    }
    [disposeSymbol]() {
        this.checkNotDisposed("The branch has already been disposed and cannot be disposed again.");
        this.disposed = true;
        this.purgeRevertibles();
        this._branch.dispose();
        for (const view of this.views) {
            view.dispose();
        }
    }
    getRemovedRoots() {
        this.assertNoUntrackedRoots();
        const trees = [];
        const cursor = this.forest.allocateCursor("getRemovedRoots");
        for (const { id, root } of this.removedRoots.entries()) {
            const parentField = this.removedRoots.toFieldKey(root);
            this.forest.moveCursorToPath({ parent: undefined, parentField, parentIndex: 0 }, cursor);
            const tree = jsonableTreeFromCursor(cursor);
            // This method is used for tree consistency comparison.
            const { major, minor } = id;
            const finalizedMajor = major !== undefined ? this.revisionTagCodec.encode(major) : major;
            trees.push([finalizedMajor, minor, tree]);
        }
        cursor.free();
        return trees;
    }
    /**
     * This sets the tip revision as the latest relevant revision for any removed roots that are loaded from a summary.
     * This needs to be called right after loading {@link this.removedRoots} from a summary to allow loaded data to be garbage collected.
     */
    setTipRevisionForLoadedData(revision) {
        this.removedRoots.setRevisionsForLoadedData(revision);
    }
    purgeRevertibles() {
        for (const revertible of this.revertibles) {
            revertible.dispose();
        }
    }
    disposeRevertible(revertible, revision) {
        this.revertibleCommitBranches.get(revision)?.dispose();
        this.revertibleCommitBranches.delete(revision);
        this.revertibles.delete(revertible);
    }
    revertRevertible(revision, kind) {
        if (this._branch.isTransacting()) {
            throw new UsageError("Undo is not yet supported during transactions.");
        }
        const revertibleBranch = this.revertibleCommitBranches.get(revision);
        assert(revertibleBranch !== undefined, 0x7cc /* expected to find a revertible commit */);
        const commitToRevert = revertibleBranch.getHead();
        const revisionForInvert = this.mintRevisionTag();
        let change = tagChange(this.changeFamily.rebaser.invert(commitToRevert, false, revisionForInvert), revisionForInvert);
        const headCommit = this._branch.getHead();
        // Rebase the inverted change onto any commits that occurred after the undoable commits.
        if (commitToRevert !== headCommit) {
            change = tagChange(rebaseChange(this.changeFamily.rebaser, change, commitToRevert, headCommit, this.mintRevisionTag).change, revisionForInvert);
        }
        this._branch.apply(change, kind === CommitKind.Default || kind === CommitKind.Redo
            ? CommitKind.Undo
            : CommitKind.Redo);
        // Derive some stats about the reversion to return to the caller.
        let revertAge = 0;
        let currentCommit = headCommit;
        while (commitToRevert.revision !== currentCommit.revision) {
            revertAge++;
            const parentCommit = currentCommit.parent;
            assert(parentCommit !== undefined, 0x9a9 /* expected to find a parent commit */);
            currentCommit = parentCommit;
        }
        return { age: revertAge };
    }
    assertNoUntrackedRoots() {
        const cursor = this.forest.getCursorAboveDetachedFields();
        const rootFields = new Set([rootFieldKey]);
        for (const { root } of this.removedRoots.entries()) {
            rootFields.add(this.removedRoots.toFieldKey(root));
        }
        if (!cursor.firstField()) {
            return;
        }
        do {
            const field = cursor.getFieldKey();
            assert(rootFields.has(field), 0xa22 /* Forest has a root field which is unknown to the detached field index */);
            rootFields.delete(field);
        } while (cursor.nextField());
    }
    /**
     * `true` iff the given branch change event is due to a remote change
     */
    isRemoteChangeEvent(event) {
        return (
        // remote changes are only ever applied to the main branch
        !this.isBranch &&
            // remote changes are applied to the main branch by rebasing it onto the trunk,
            // no other rebases are allowed on the main branch so this means any replaces that are not
            // transaction commits are remote changes
            event.type === "replace" &&
            getChangeReplaceType(event) !== "transactionCommit");
    }
}
/**
 * The name of the telemetry event logged for calls to {@link TreeCheckout.revertRevertible}.
 * @privateRemarks Exposed for testing purposes.
 */
TreeCheckout.revertTelemetryEventName = "RevertRevertible";
/**
 * Run a synchronous transaction on the given shared tree view.
 * This is a convenience helper around the {@link SharedTreeFork#transaction} APIs.
 * @param view - the view on which to run the transaction
 * @param transaction - the transaction function. This will be executed immediately. It is passed `view` as an argument for convenience.
 * If this function returns an `Abort` result then the transaction will be aborted. Otherwise, it will be committed.
 * @returns whether or not the transaction was committed or aborted
 */
export function runSynchronous(view, transaction) {
    view.transaction.start();
    const result = transaction(view);
    return result === TransactionResult.Abort
        ? view.transaction.abort()
        : view.transaction.commit();
}
//# sourceMappingURL=treeCheckout.js.map